/*
LINKFILE.C

C version of various pedigree and locus handling functions, swiped
for DOLINK so that locus input/output routines can be used,
but also used by MFLINK and other progs
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "linkfile.h"

#ifdef USEDCERROR
#include "dcerror.hpp"
#define error(s,t) error_func(__LINE__,__FILE__,s,t)
error_func(unsigned l,char *f,char *s1,char *s2)
{
dcerror(99,"Error on line %d of source file %s:\n %s%s",l,f,s1,s2);
return 0;
}
#else
error_func(unsigned l,char *f,char *s1,char *s2)
{
fprintf(stderr,"Error on line %d of source file %s:\n %s%s\n\n",l,f,s1,s2);
exit(1);
return 0;
}
#endif

int locus_def_output(FILE *fp,dc_locus_def *def,int sexlinked)
{
int i,j;
fprintf(fp,"%d %d << type, number of alleles\n",def->type,def->nalls);
for (i=0;i<def->nalls;++i)
   fprintf(fp,"%f ",def->freq[i]);
fprintf(fp,"\n");
switch (def->type)
  {
  case DCL_AFF:
    fprintf(fp,"%d",def->info.aff.nliab);
    for (i=0;i<def->info.aff.nliab;++i)
      {
      fprintf(fp,"%f %f %f",
           def->info.aff.pen[i][0],
           def->info.aff.pen[i][1],
           def->info.aff.pen[i][2]);
      if (sexlinked)
          fprintf(fp,"%f %f",
           def->info.aff.male_pen[i][0],
           def->info.aff.male_pen[i][1]);
      }
    break;
  case DCL_BINARY:
    fprintf(fp,"%d\n",def->info.bin.nfact);
    for (i=0;i<def->nalls;++i)
      {
      for (j=0;j<def->info.bin.nfact;++j)
        fprintf(fp,"%c ",def->info.bin.fact[i][j]);
      fprintf(fp,"\n");
      }
    break;
  case DCL_QUAN:
    fprintf(fp,"%f %f",def->info.quan.mAA,&def->info.quan.maa);
    fprintf(fp,"%f",def->info.quan.dom);
    fprintf(fp,"%f %f",def->info.quan.var,def->info.quan.hetmult);
        break;
  case DCL_NUMBERED:
    break;
  }
if (def->isrisk) fprintf(fp,"%d\n",def->riskall);
fprintf(fp,"\n");
return 1;
}

int locus_def_input(FILE *fp,dc_locus_def *def,int sexlinked,int risklocus)
{
char line[1000],rest[1000];
int i,j;
do {
   if (!fgets(line,1000,fp)) return 0;
   } while (sscanf(line,"%d %d %[^\n]",&def->type,&def->nalls,rest)<2);
if (!fgets(line,1000,fp)) return 0;
def->isrisk=risklocus;
if ((def->type==DCL_AFF||def->type==DCL_QUAN) && def->nalls!=2)
  { error("Sorry, can only handle affection and quantitative loci with two alleles",""); return 0; }
for (i=0;i<def->nalls;++i,strcpy(line,rest),*rest='\0')
   if (sscanf(line,"%f %[^\n]",&def->freq[i],rest)<1 ||
         def->freq[i]>1.0)  /* load floating point */
   { error("Problem reading allele frequencies",""); return 0; }
switch (def->type)
  {
  case DCL_AFF:
    if (!fgets(line,1000,fp) ||
          sscanf(line,"%d",&def->info.aff.nliab)!=1) return 0;
    for (i=0;i<def->info.aff.nliab;++i)
      {
      if (!fgets(line,1000,fp) ||
          sscanf(line,"%f %f %f",
           &def->info.aff.pen[i][0],
           &def->info.aff.pen[i][1],
           &def->info.aff.pen[i][2])!=3) return 0;
      if (sexlinked && (!fgets(line,1000,fp) ||
          sscanf(line,"%f %f",
           &def->info.aff.male_pen[i][0],
           &def->info.aff.male_pen[i][1])!=2)) return 0;
      }
    break;
  case DCL_BINARY:
    if (!fgets(line,1000,fp) ||
          sscanf(line,"%d",&def->info.bin.nfact)!=1) return 0;
    for (i=0;i<def->nalls;++i)
      {
      if (!fgets(line,1000,fp)) return 0;
      for (j=0;j<def->info.bin.nfact;++j,strcpy(line,rest))
        if (sscanf(line,"%c %[^\n]",&def->info.bin.fact[i][j],rest)<1) return 0;
      }
    break;
  case DCL_QUAN:
    if (!fgets(line,1000,fp) ||
          sscanf(line,"%f %f",&def->info.quan.mAA,&def->info.quan.maa)!=2) return 0;
    if (!fgets(line,1000,fp) ||
          sscanf(line,"%f",&def->info.quan.dom)!=1) return 0;
    if (!fgets(line,1000,fp) ||
          sscanf(line,"%f %f",&def->info.quan.var,&def->info.quan.hetmult)!=2) return 0;
        break;
  case DCL_NUMBERED:
    break;
  }
if (def->isrisk&&(!fgets(line,1000,fp)||sscanf(line,"%d",&def->riskall)!=1))
  return 0;
return 1;
}

read_person(FILE *fp,dc_person *ind,dc_locus_def *def,int nloci)
{
int i;
char line[1001],rest[1000],*ptr;
ind->nspouse=0;
ind->parents=NULL;
rest[0]='\0';
if (!fgets(line,1000,fp) || sscanf(line,"%s %d %d %d %*d %*d %*d %d %d %[^\n]",
  &ind->fname,&ind->idnum,&ind->fidnum,&ind->midnum,
  &ind->sex,&ind->isproband,rest)!=7) return 0;
ptr=rest;
for (i=0;i<nloci;++i)
  if (!locus_input(&ind->locus[i],&def[i],&ptr)) return 0;
return 1;
}

read_pedigree(FILE *fp,dc_person *ped,dc_locus_def *def,int nloci)
{
static dc_person last_one;
int i;
if (last_one.idnum==0)
  if (!read_person(fp,&last_one,def,nloci)) return 0;
ped[0]=last_one;
i=1;
/* set last_one.idnum back to 0 if read fails */
while ((read_person(fp,&last_one,def,nloci)||(last_one.idnum=0)) &&
         !strcmp(ped[i-1].fname,last_one.fname))
        ped[i++]=last_one;
return i;
}

set_parents(dc_person *child,dc_person *ped,int count)
{
dc_mating *mating;
int i,j,k;
for (i=0;i<count;++i)
  if (ped[i].idnum==child->fidnum) break;
if (i==count) { error("Father not found",""); return 0; }
for (j=0;j<ped[i].nspouse;++i)
  if (ped[i].spouse[j]->mother->idnum==child->midnum) break;
if (j==ped[i].nspouse) /* mating not already found */
  {
  for (k=0;k<count;++k)
    if (ped[k].idnum==child->midnum) break;
  if (k==count) { error("Mother not found",""); return 0; }
  mating=(dc_mating*)malloc(sizeof(dc_mating));
  if (mating==NULL) { error("No memory creating mating",""); return 0; }
  mating->nchild=0;
  mating->father=&ped[i];
  mating->mother=&ped[k];
  ped[i].spouse[ped[i].nspouse++]=mating;
  ped[k].spouse[ped[k].nspouse++]=mating;
  }
child->parents=ped[i].spouse[j];
child->parents->child[child->parents->nchild++]=child;
return 1;
}

free_matings(dc_person *ped,int count)
{
int i;
for (i=0;i<count;++i)
  if (ped[i].sex==1)
    while (ped[i].nspouse)
      free(ped[i].spouse[--ped[i].nspouse]);
return 1;
}

make_matings(dc_person *ped,int count)
{
int i;
for (i=0;i<count;++i)
  if (ped[i].fidnum!=0 && set_parents(&ped[i],ped,count)==0) return 0;
return 1;
}

int locus_input(dc_locus *locus,dc_locus_def *def,char **line)
{
int nread,n,i;
locus->type=def->type;
switch (def->type)
  {
  case DCL_AFF:
    if (def->info.aff.nliab==1)
      {
      if (sscanf(*line," %d%n",&locus->data.aff[1],&nread)!=1) return 0;
      }
    else
      if (sscanf(*line," %d %d%n",&locus->data.aff[0],
            &locus->data.aff[1],&nread)!=2) return 0;
    break;
  case DCL_QUAN:
    if (sscanf(*line," %f%n",&locus->data.qval,&nread)!=1) return 0;
    break;
  case DCL_BINARY:
    for (i=0,nread=0;i<def->info.bin.nfact;++i,nread+=n)
      if (sscanf(*line," %c%n",&locus->data.bin[i],&n)!=1) return 0;
    break;
  case DCL_NUMBERED:
      if (sscanf(*line," %d %d%n",&locus->data.all[0],
            &locus->data.all[1],&nread)!=2) return 0;
    break;
  }    
*line+=nread;
return 1;
}

read_data_file(FILE *fp,dc_locus_def *def,int *sexlinked,int *risklocus)
{
int nloci,i;
char line[1001];
if (!fgets(line,1000,fp) || 
      sscanf(line,"%d %d %d",&nloci,risklocus,sexlinked)!=3 ||
      !fgets(line,1000,fp) || !fgets(line,1000,fp))
        { error("Bad datafile",""); return 0; }
for (i=0;i<nloci;++i)
   if (!locus_def_input(fp,&def[i],*sexlinked,*risklocus==i+1)) return 0;
return nloci;
}

